@@ -10,28 +10,52 @@ sub new
   my ($class, $asp) = @_;
   
   my $s = bless {
-    asp => $asp,
-    q => $asp->{q},
-    r => $asp->{r},
+#    asp => $asp,
+#    'q' => $asp->{'q'},
+#    r => $asp->{r},
   }, $class;
   
+  # Taken from CGI::Util by Lincoln Stein:
+  $s->{_unescaper} = sub {
+    my $todecode = shift;
+    return undef unless defined($todecode);
+    $todecode =~ tr/+/ /;       # pluses become spaces
+    $todecode =~ s/%(?:([0-9a-fA-F]{2})|u([0-9a-fA-F]{4}))/
+    defined($1)? chr hex($1) : utf8_chr(hex($2))/ge;
+    return $todecode;
+  };
+  
   {
     no warnings 'uninitialized';
     $s->{cookies} = {
-      map { 
+      map {
         my ($k,$v) = split /\=/, $_;
         chomp($k);
-        ($k => $asp->{q}->unescape($v) )
+        $k =~ s/^\s+//;
+        ($k => $v . '' )
       } split /;/, $ENV{HTTP_COOKIE}
     };
   }
   
-  foreach my $cookie ( keys(%{ $s->{cookies} }) )
+  while( my ($key,$data) = each( %{ $s->{cookies} } ) )
   {
-    next unless $s->{cookies}->{$cookie} =~ m/\=/;
-    my %vals = map { my($k,$v) = split /\=/, $_; chomp($k); ($k => $s->{q}->unescape($v)) } split /&/, $s->{cookies}->{$cookie};
-    $s->{cookies}->{$cookie} = \%vals;
-  }# end foreach()
+    if( $data =~ m/\%3D/i )
+    {
+      $data = $s->{_unescaper}->( $data );
+      my %info = map {
+        my ($k,$v) = split /\=/, $_;
+        chomp($k);
+        $key =~ s/^\s+//;
+        ( $k => $v )
+      } split /&/, $data;
+      $s->{cookies}->{$key} = \%info;
+    }
+    else
+    {
+      $data = $s->{_unescaper}->( $data );
+      $s->{cookies}->{$key} = $data;
+    }# end if()
+  }# end while()
   
   return $s;
 }# end new()
@@ -41,17 +65,26 @@ sub new
 sub Cookies
 {
   my ($s, $name, $key ) = @_;
-  
-  return unless exists($s->{cookies}->{$name});
-  
-  if( defined($key) )
-  {
-    return $s->{cookies}->{$name}->{$key};
-  }
-  else
-  {
-    return $s->{cookies}->{$name};
-  }# end if()
+	
+	if( @_ == 1 )
+	{
+		return keys( %{ $s->{cookies} } );
+	}
+	elsif( @_ == 2 )
+	{
+		return unless exists( $s->{cookies}->{$name} );
+		return $s->{cookies}->{$name};
+	}
+	elsif( @_ == 3 )
+	{
+		return unless defined($name) && defined($key);
+		return unless exists( $s->{cookies}->{$name} ) && exists( $s->{cookies}->{$name}->{$key} );
+		return $s->{cookies}->{$name}->{$key};
+	}
+	else
+	{
+		return;
+	}# end if()
 }# end Cookies()
 
 
@@ -59,16 +92,48 @@ sub Cookies
 sub Form
 {
   my $s = shift;
+  
   if( @_ )
   {
     my $arg = shift;
-    my $val = $s->{q}->param( $arg );
-    return defined($val) ? $val : undef;
+    my $val = $s->_q->param( $arg );
+    if( defined($val) )
+    {
+      return $val;
+    }
+    else
+    {
+      if( my $page_args = $s->asp->session->{__lastArgs} )
+      {
+        return $page_args->{ $arg };
+      }# end if()
+    }# end if()
   }
   else
   {
-    my %info = map { $_ => $s->{q}->param( $_ ) } $s->{q}->param;
-    return \%info;
+    return $s->{_form} if $s->{_form};
+    no warnings 'uninitialized';
+    my $page_args = { };
+    if( my $last = $s->asp->session->{__lastArgs} )
+    {
+      $page_args = ref($last) ? $last : { };
+    }# end if()
+    
+    my %info = %$page_args;
+    foreach my $field ( $s->_q->param )
+    {
+      my @data = $s->_q->param( $field );
+      if( scalar(@data) > 1 )
+      {
+        $info{$field} = [ @data ];
+      }
+      else
+      {
+        $info{$field} = shift(@data);
+      }# end if()
+    }# end foreach()
+    
+    return $s->{_form} = \%info;
   }# end if()
 }# end Form()
 
@@ -78,16 +143,33 @@ sub FileUpload
 {
   my ($s, $field, $arg) = @_;
   
-  my $ifh = $s->{q}->upload($field);
-  my $upInfo = $s->{q}->uploadInfo( $ifh );
-  no warnings 'uninitialized';
-  my %info = (
-    ContentType           => $upInfo->{'Content-Type'},
-    FileHandle            => $ifh,
-    BrowserFile           => $s->Form->{ $field } . "",
-    'Content-Disposition' => $upInfo->{'Content-Disposition'},
-    'Mime-Header'         => $upInfo->{type},
-  );
+  my $ifh = $s->_q->upload($field);
+  my %info = ();
+  my $upInfo = { };
+  
+  if( $s->_q->isa('CGI::Simple') || $s->_q->isa('Apache2::ASP::SimpleCGI') )
+  {
+    no warnings 'uninitialized';
+    %info = (
+      ContentType           => $s->_q->upload_info( $field, 'mime' ),
+      FileHandle            => $ifh,
+      BrowserFile           => $s->Form->{ $field } . "",
+      'Content-Disposition' => 'attachment',
+      'Mime-Header'         => $s->_q->upload_info( $field, 'mime' ),
+    );
+  }
+  else
+  {
+    $upInfo = $s->_q->uploadInfo( $ifh );
+    no warnings 'uninitialized';
+    %info = (
+      ContentType           => $upInfo->{'Content-Type'},
+      FileHandle            => $ifh,
+      BrowserFile           => $s->Form->{ $field } . "",
+      'Content-Disposition' => $upInfo->{'Content-Disposition'},
+      'Mime-Header'         => $upInfo->{type},
+    );
+  }# end if()
   
   if( $field )
   {
@@ -118,7 +200,20 @@ sub FileUpload
 #==============================================================================
 sub QueryString
 {
-  $ENV{HTTP_QUERYSTRING};
+  my ($s) = shift;
+  if( my $field = shift(@_) )
+  {
+    my @parts = split /\&/, $ENV{HTTP_QUERYSTRING};
+    my %info = map {
+      my ($k,$v) = split /\=/, $_;
+      $s->{_unescaper}->($k) => $s->{_unescaper}->($v)
+    } @parts;
+    return $info{$field};
+  }
+  else
+  {
+    return $ENV{HTTP_QUERYSTRING};
+  }# end if()
 }# end Form()
 
 
@@ -138,9 +233,38 @@ sub ServerVariables
 }# end ServerVariables()
 
 
+#==============================================================================
+# Be Lazy - only parse the DOM if asked:
+# Secret functionality until finalized - do not use!
+sub Document
+{
+  my $s = shift;
+  
+  return $s->{_document} if $s->{_document};
+  
+  # We haven't parsed it yet - do it now:
+  my $file = $s->ServerVariables('SCRIPT_FILENAME');
+  open my $ifh, '<', $file
+    or die "Cannot open '$file': $!";
+  local $/ = undef;
+  my $txt = <$ifh>;
+  close($ifh);
+  require Apache2::ASP::DOM::Parser;
+  $s->{_document} = Apache2::ASP::DOM::Parser->parse( $txt );
+}# end Document()
+
+
+#==============================================================================
+sub asp { $main::_ASP::ASP }
+sub _q { $main::_ASP::ASP->{'q'} }
+sub _r { $main::_ASP::ASP->{'r'} }
+
+
 #==============================================================================
 sub DESTROY
 {
+  my $s = shift;
+  delete($s->{$_}) foreach keys(%$s);
 }# end DESTROY()
 
 1;# return true:
@@ -240,6 +364,13 @@ Called without a C<$key>, returns a sorted list of all keys in C<%ENV>.
 
 Called with a C<$key>, returns the value associated with that element in C<%ENV>.
 
+=head2 Document( )
+
+Returns the L<Apache2::ASP::DOM::Document> of the current ASP script.
+
+B<NOTE>: If called from within an Included file, the Document will only contain the
+DOM for that file, not the entire document.
+
 =head1 BUGS
 
 It's possible that some bugs have found their way into this release.
@@ -248,7 +379,7 @@ Use RT L<http://rt.cpan.org/NoAuth/Bugs.html?Dist=Apache2-ASP> to submit bug rep
 
 =head1 HOMEPAGE
 
-Please visit the Apache2::ASP homepage at L<http://apache2-asp.no-ip.org/> to see examples
+Please visit the Apache2::ASP homepage at L<http://www.devstack.com/> to see examples
 of Apache2::ASP in action.
 
 =head1 AUTHOR
